@@ -18,12 +18,18 @@ class ContainerLauncher:
 
     __ASG_DRAINING_TIMEOUT_IN_MINUTES = 5
     __INSTANCE_TERMINATION_TIMEOUT_IN_MINUTES = 5
+    __SPOT_INSTANCE_TERMINATION_TIMEOUT_IN_MINUTES = 25
     __ECS_CLUSTER_ATTACHMENTS_UPDATE_TIMEOUT_IN_MINUTES = 5
     __ASG_FINISHING_SCALING_ACTIVITIES_TIMEOUT_IN_MINUTES = 5
     __EC2_INSTANCE_DEPLOYING_TIMEOUT_IN_MINUTES = 5
     __EC2_INSTANCE_REGISTRATION_TIMEOUT_IN_MINUTES = 5
     __EC2_INSTANCE_UNUSED_MEMORY = 512
     __ECS_TASK_LOG_STREAM_PREFIX = 'acl-task'
+    __ECS_TASK_EXECUTION_POLICY_ARN = 'arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy'
+    __IAM_EC2_MANAGED_POLICY_ARNS = [
+        'arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role',
+        #'arn:aws:iam::aws:policy/AmazonSSMManagedInstanceCore'
+    ]
 
     def __init__(self, options=None):
         self.__options = None
@@ -81,8 +87,16 @@ class ContainerLauncher:
 
     def start_container(self, container_input):
         entry_id = container_input.get('id', str(uuid.uuid4()).replace('-', ''))
-        ec2_sg_id = container_input.get('ec2_sg_id')
-
+        ec2_sg_ids = container_input.get('ec2_sg_ids')
+        iam_ecs_task_policy_documents = None
+        iam_ec2_policy_documents = None
+        if container_input.get('iam') is not None:
+            if container_input['iam'].get('ecs_task') is not None:
+                if container_input['iam']['ecs_task'].get('policy_documents'):
+                    iam_ecs_task_policy_documents = container_input['iam']['ecs_task']['policy_documents']
+            if container_input['iam'].get('ec2') is not None:
+                if container_input['iam']['ec2'].get('policy_documents'):
+                    iam_ec2_policy_documents = container_input['iam']['ec2']['policy_documents']
         try:
             self.__options['name_suffix'] = container_input.get('name_suffix', str(uuid.uuid4()).replace('-', ''))
 
@@ -136,8 +150,10 @@ class ContainerLauncher:
             }).get_image_id({'instance_type': instance_type})
             subnet_ids = container_input['subnet_ids']
 
-            ec2_iam_role_arn = self.__create_ec2_iam_role()
-            self.__create_ecs_task_iam_role()
+            ec2_iam_role_arn = self.__create_ec2_iam_role(iam_ec2_policy_documents)
+
+            self.__create_ecs_task_iam_role(iam_ecs_task_policy_documents)
+            self.__create_ecs_task_execution_iam_role()
             account_id = ec2_iam_role_arn.split(':')[4]
             self.__create_instance_profile()
             self.__add_role_to_instance_profile()
@@ -147,7 +163,7 @@ class ContainerLauncher:
                 'instance_type': instance_type,
                 'storages': storages,
                 'is_spot': is_spot,
-                'ec2_sg_id': ec2_sg_id
+                'ec2_sg_ids': ec2_sg_ids
             })
             self.__create_asg({
                 'subnet_ids': subnet_ids,
@@ -160,7 +176,7 @@ class ContainerLauncher:
                 self.__add_scale_in_policy({})
             self.__create_capacity_provider()
             self.__create_ecs_cluster()
-            self.__wait_for_instance_deploying()
+            self.__wait_for_instance_deploying(is_spot)
             self.__wait_for_instance_registration_in_ecs()
 
             memory = self.__get_available_memory()
@@ -282,21 +298,14 @@ class ContainerLauncher:
 
         self.__resources['instance_profile_arn'] = instance_profile_arn
 
-        # waiter = self.clients['iam'].get_waiter('instance_profile_exists')
-
-        # waiter.wait(
-        #    InstanceProfileName=instance_profile_name
-        # )
-
         # Sleep to make the instance profile available
         time.sleep(15)
 
-    def __create_ec2_iam_role(self):
+    def __create_ec2_iam_role(self, policy_documents=None):
         name_suffix = self.__options['name_suffix']
         role_name = f'{self.__name_prefix}-ec2-role-{name_suffix}'
-        policy_name = f'{self.__name_prefix}-ec2-policy-{name_suffix}'
 
-        assume_role_policy_document = json.dumps({
+        assume_role_policy_document = {
             "Statement": [
                 {
                     "Effect": "Allow",
@@ -306,11 +315,11 @@ class ContainerLauncher:
                     "Action": "sts:AssumeRole"
                 }
             ]
-        })
+        }
 
         create_role_response = self.clients['iam'].create_role(
             RoleName=role_name,
-            AssumeRolePolicyDocument=assume_role_policy_document,
+            AssumeRolePolicyDocument=json.dumps(assume_role_policy_document),
             Tags=self.__tags
         )
 
@@ -318,28 +327,32 @@ class ContainerLauncher:
 
         self.logger.debug(f'The EC2 IAM role {role_name} has been created')
 
-        self.clients['iam'].put_role_policy(
-            PolicyDocument='{"Version":"2012-10-17","Statement":{"Effect":"Allow","Action":"*","Resource":"*"}}',
-            PolicyName=policy_name,
-            RoleName=role_name,
-        )
+        for managed_policy_arn in self.__IAM_EC2_MANAGED_POLICY_ARNS:
+            self.clients['iam'].attach_role_policy(
+                RoleName=role_name,
+                PolicyArn=managed_policy_arn
+            )
 
-        self.__resources['ec2_iam_policy_name'] = policy_name
+        self.__resources['ec2_iam_role_policy_document_names'] = []
 
-        # waiter = self.clients['iam'].get_waiter('role_exists')
-        # waiter.wait(
-        #    RoleName=role_name,
-        # )
+        if policy_documents is not None:
+            for i, policy_document in enumerate(policy_documents):
+                policy_name = f'{self.__name_prefix}-ec2-policy-{i}-{name_suffix}'
+                self.__resources['ec2_iam_role_policy_document_names'].append(policy_name)
+                self.clients['iam'].put_role_policy(
+                    PolicyDocument=json.dumps(policy_document),
+                    PolicyName=policy_name,
+                    RoleName=role_name
+                )
 
         # Sleep to make the IAM role available
         time.sleep(15)
 
         return create_role_response['Role']['Arn']
 
-    def __create_ecs_task_iam_role(self):
+    def __create_ecs_task_iam_role(self, policy_documents):
         name_suffix = self.__options['name_suffix']
         role_name = f'{self.__name_prefix}-ecs-task-role-{name_suffix}'
-        policy_name = f'{self.__name_prefix}-ecs-task-policy-{name_suffix}'
 
         assume_role_policy_document = json.dumps({
             "Statement": [
@@ -363,17 +376,52 @@ class ContainerLauncher:
 
         self.logger.debug(f'The ECS task IAM role {role_name} has been created')
 
-        self.clients['iam'].put_role_policy(
-            PolicyDocument='{"Version":"2012-10-17","Statement":{"Effect":"Allow","Action":"*","Resource":"*"}}',
-            PolicyName=policy_name,
+        self.__resources['ecs_task_iam_role_policy_document_names'] = []
+
+        if policy_documents is not None:
+            for i, policy_document in enumerate(policy_documents):
+                policy_name = f'{self.__name_prefix}-ecs-task-policy-{i}-{name_suffix}'
+                self.__resources['ecs_task_iam_role_policy_document_names'].append(policy_name)
+                self.clients['iam'].put_role_policy(
+                    PolicyDocument=json.dumps(policy_document),
+                    PolicyName=policy_name,
+                    RoleName=role_name
+                )
+
+        # Sleep to make the IAM role available
+        time.sleep(15)
+
+        return create_role_response['Role']['Arn']
+
+    def __create_ecs_task_execution_iam_role(self):
+        name_suffix = self.__options['name_suffix']
+        role_name = f'{self.__name_prefix}-ecs-task-execution-role-{name_suffix}'
+
+        assume_role_policy_document = json.dumps({
+            "Statement": [
+                {
+                    "Effect": "Allow",
+                    "Principal": {
+                        "Service": "ecs-tasks.amazonaws.com"
+                    },
+                    "Action": "sts:AssumeRole"
+                }
+            ]
+        })
+
+        create_role_response = self.clients['iam'].create_role(
             RoleName=role_name,
+            AssumeRolePolicyDocument=assume_role_policy_document,
+            Tags=self.__tags
         )
 
-        self.__resources['ecs_task_iam_policy_name'] = policy_name
+        self.__resources['ecs_task_execution_iam_role_name'] = role_name
 
-        waiter = self.clients['iam'].get_waiter('role_exists')
-        waiter.wait(
+        self.logger.debug(f'The ECS task execution IAM role {role_name} has been created')
+
+        self.clients['iam'].attach_role_policy(
             RoleName=role_name,
+            PolicyArn=self.__ECS_TASK_EXECUTION_POLICY_ARN
         )
 
         # Sleep to make the IAM role available
@@ -393,12 +441,19 @@ class ContainerLauncher:
         except Exception as e:
             self.logger.warning(f'An error occurred during the remove_role_from_instance_profile operation: {str(e)}')
         try:
-            if self.__resources.get('ec2_iam_role_name') and self.__resources.get('ec2_iam_policy_name'):
-                self.clients['iam'].delete_role_policy(
-                    RoleName=self.__resources['ec2_iam_role_name'],
-                    PolicyName=self.__resources['ec2_iam_policy_name']
+            for managed_policy_arn in self.__IAM_EC2_MANAGED_POLICY_ARNS:
+                self.clients['iam'].detach_role_policy(
+                    RoleName=self.__resources.get('ec2_iam_role_name'),
+                    PolicyArn=managed_policy_arn
                 )
-                self.logger.debug(f"The IAM policy {self.__resources['ec2_iam_policy_name']} has been deleted")
+            ec2_iam_role_policy_document_names = self.__resources['ec2_iam_role_policy_document_names']
+            if self.__resources.get('ec2_iam_role_name') and ec2_iam_role_policy_document_names and len(ec2_iam_role_policy_document_names) > 0:
+                for ec2_iam_role_policy_document_name in ec2_iam_role_policy_document_names:
+                    self.clients['iam'].delete_role_policy(
+                        RoleName=self.__resources['ec2_iam_role_name'],
+                        PolicyName=ec2_iam_role_policy_document_name
+                    )
+                self.logger.debug(f"The IAM policies {str(self.__resources['ec2_iam_role_policy_document_names'])} has been deleted")
         except Exception as e:
             self.logger.warning(f'An error occurred during the delete_role_policy operation: {str(e)}')
         try:
@@ -406,7 +461,7 @@ class ContainerLauncher:
                 self.clients['iam'].delete_role(RoleName=self.__resources['ec2_iam_role_name'])
                 self.logger.debug(f"EC2 IAM role {self.__resources['ec2_iam_role_name']} has been deleted")
         except Exception as e:
-            self.logger.warning(f'An error occurred during the delete_role operation: {str(e)}')
+            self.logger.warning(f"An error occurred during the {self.__resources['ec2_iam_role_name']} delete_role operation: {str(e)}")
         try:
             if self.__resources.get('instance_profile_name'):
                 self.clients['iam'].delete_instance_profile(
@@ -442,7 +497,7 @@ class ContainerLauncher:
 
                 self.logger.debug(f"Set DesiredCapacity=0 for {asg_name} ASG")
 
-                self.__wait_for_instance_terminating(asg_name)
+                self.__wait_for_instance_terminating(asg_name, )
 
                 self.__wait_for_draining_asg(asg_name)
 
@@ -492,9 +547,13 @@ class ContainerLauncher:
         except Exception as e:
             self.logger.warning(f'An error occurred during the deregister_task_definition operation: {str(e)}')
         try:
-            if self.__resources.get('ecs_task_iam_policy_name'):
-                self.clients['iam'].delete_role_policy(RoleName=self.__resources.get('ecs_task_iam_role_name'),
-                                                       PolicyName=self.__resources.get('ecs_task_iam_policy_name'))
+            policy_document_names = self.__resources.get('ecs_task_iam_role_policy_document_names')
+            if self.__resources.get('ecs_task_iam_role_name') and policy_document_names:
+                for policy_document_name in policy_document_names:
+                    self.clients['iam'].delete_role_policy(
+                        RoleName=self.__resources.get('ecs_task_iam_role_name'),
+                        PolicyName=policy_document_name
+                    )
                 self.logger.debug(
                     f"The ECS task IAM policy {self.__resources.get('ecs_task_iam_policy_name')} has been deleted")
         except Exception as e:
@@ -505,7 +564,19 @@ class ContainerLauncher:
                 self.logger.debug(
                     f"The ECS task IAM role {self.__resources.get('ecs_task_iam_role_name')} has been deleted")
         except Exception as e:
-            self.logger.warning(f'An error occurred during the delete_role operation: {str(e)}')
+            self.logger.warning(f"An error occurred during the {self.__resources.get('ecs_task_iam_role_name')} delete_role operation: {str(e)}")
+        try:
+            ecs_task_execution_iam_role_name = self.__resources.get('ecs_task_execution_iam_role_name')
+            if ecs_task_execution_iam_role_name:
+                self.clients['iam'].detach_role_policy(
+                    RoleName=ecs_task_execution_iam_role_name,
+                    PolicyArn=self.__ECS_TASK_EXECUTION_POLICY_ARN
+                )
+                self.clients['iam'].delete_role(RoleName=ecs_task_execution_iam_role_name)
+                self.logger.debug(
+                    f"The ECS task execution IAM role {ecs_task_execution_iam_role_name} has been deleted")
+        except Exception as e:
+            self.logger.warning(f"An error occurred during the {self.__resources.get('ecs_task_execution_iam_role_name')} delete_role operation: {str(e)}")
 
     def __create_asg(self, asg_opts):
         name_suffix = self.__options['name_suffix']
@@ -530,7 +601,6 @@ class ContainerLauncher:
 
         self.clients['autoscaling'].create_auto_scaling_group(
             AutoScalingGroupName=asg_name,
-            # LaunchConfigurationName=self.__resources['launch_config_name'],
             LaunchTemplate={
                 'LaunchTemplateName': self.__resources['launch_template_name'],
             },
@@ -778,7 +848,7 @@ class ContainerLauncher:
         response = self.clients['ecs'].register_task_definition(
             family=task_definition_name,
             taskRoleArn=self.__resources['ecs_task_iam_role_name'],
-            executionRoleArn=self.__resources['ecs_task_iam_role_name'],
+            executionRoleArn=self.__resources['ecs_task_execution_iam_role_name'],
             networkMode='host',
             containerDefinitions=[
                 {
@@ -862,18 +932,20 @@ class ContainerLauncher:
 
         return task_arn
 
-    def __wait_for_instance_deploying(self):
+    def __wait_for_instance_deploying(self, is_spot):
         instances = self.__get_asg_instances()
 
         start_time = datetime.datetime.now()
 
         self.logger.info('Waiting for an EC2 instance deploying')
 
+        timeout_in_minutes = self.__SPOT_INSTANCE_TERMINATION_TIMEOUT_IN_MINUTES if is_spot else self.__INSTANCE_TERMINATION_TIMEOUT_IN_MINUTES
+
         while True:
             now_time = datetime.datetime.now()
             diff = now_time - start_time
             diff_minutes = diff.total_seconds() / 60
-            if diff_minutes > self.__EC2_INSTANCE_DEPLOYING_TIMEOUT_IN_MINUTES:
+            if diff_minutes > timeout_in_minutes:
                 raise Exception('EC2 instance deploying timeout')
             if len(instances) > 0:
                 self.logger.debug('The EC2 instance ID: ' + instances[0]['InstanceId'])
@@ -941,7 +1013,7 @@ class ContainerLauncher:
         launch_template_name = f'{self.__name_prefix}-launch-config-{name_suffix}'
         user_data = launch_template_opts.get('user_data', '')
         instance_type = launch_template_opts['instance_type']
-        ec2_sg_id = launch_template_opts['ec2_sg_id']
+        ec2_sg_ids = launch_template_opts['ec2_sg_ids']
         is_spot = launch_template_opts.get('is_spot', False)
         storages = launch_template_opts.get('storages')
 
@@ -1009,8 +1081,8 @@ mount /dev/md0 %s
             ],
         }
 
-        if ec2_sg_id is not None:
-            launch_template_data['SecurityGroupIds'] = [ec2_sg_id]
+        if ec2_sg_ids is not None and len(ec2_sg_ids) > 0:
+            launch_template_data['SecurityGroupIds'] = ec2_sg_ids
 
         if is_spot:
             launch_template_data['InstanceMarketOptions'] = {
